import { Store, select } from '@ngrx/store';

import { Host, Port, DiscoveryModeType, Service } from '@overflow/model/discovery';
import { RPCSubscriber } from '@overflow/commons/ui/decorator/RPCSubscriber';
import { RPCError } from '@overflow/rpc-js';

import { ProbeService } from '../../service/probe.service';
import { DiscoverRequestInfo } from '../model';

import * as AppStore from '../../store';
import * as DiscoveryRequestStore from '../../store/discovery/request';
import * as UserStore from '../../store/environment/user';
import * as InfraInfraStore from '../../store/infra/infra';
import { take, map } from 'rxjs/operators';
import { DiscoveryMessageType, InfraDisplayType, DiscoveryStatusType } from '../type';
import { Subject, Observable } from 'rxjs';

export class DiscoverySession {

  private _discoverRequestInfo: DiscoverRequestInfo | null;
  private _requestID: string | null;
  private startDate: Date | null;
  private discoverySubject: Subject<{ messageType: DiscoveryMessageType, params: any[] }> | null;

  private constructor(
    private store: Store<any>,
    private probeService: ProbeService,
  ) {
  }

  public static requestDiscover(store: Store<any>, probeService: ProbeService): void {
    store.pipe(
      take(1),
      select((state) => {
        const memberID = AppStore.EnvironmentSelector.UserSelector.selectMemberID(state);
        const zone = AppStore.DiscoverySelector.ConfigSelector.selectZone(state);
        const host = AppStore.DiscoverySelector.ConfigSelector.selectHost(state);
        const port = AppStore.DiscoverySelector.ConfigSelector.selectPort(state);
        const discoverHost = AppStore.DiscoverySelector.ConfigSelector.selectDiscoverHost(state);
        const discoverPort = AppStore.DiscoverySelector.ConfigSelector.selectDiscoverPort(state);
        const discoverService = AppStore.DiscoverySelector.ConfigSelector.selectDiscoverService(state);
        const discoverySession = AppStore.DiscoverySelector.RequestSelector.selectDiscoverySession(state);

        if (null !== discoverySession) {
          return null;
        }
        const discoverRequestInfo = {
          requesterID: memberID,
          zone,
          host,
          port,
          discoverHost,
          discoverPort,
          discoverService,
        };
        const _discoverySession = new DiscoverySession(store, probeService);
        _discoverySession.init(discoverRequestInfo);

        return _discoverySession;
      }),
    ).pipe(
      map((discoverySession) => {
        if (null === discoverySession) {
          return;
        }

        store.dispatch(new UserStore.ChangeShowWelcomPage({ showWelcomPage: false }));
        store.dispatch(new InfraInfraStore.ChangeInfraDisplayType({ infraDisplayType: InfraDisplayType.MAP }));

        store.dispatch(new DiscoveryRequestStore.RequestDiscover({ discoverySession }));
        // store.dispatch(new UserStore.ChangeShowWelcomPage({ showWelcomPage: false }));
        // store.dispatch(new InfraInfraStore.ChangeInfraDisplayType({ infraDisplayType: InfraDisplayType.MAP }));
      }),
    ).subscribe();
  }

  public static requestDiscoverStop(store: Store<any>): void {
    store.pipe(
      take(1),
      select((state) => {
        const discoverySession = AppStore.DiscoverySelector.RequestSelector.selectDiscoverySession(state);

        if (null === discoverySession) {
          return null;
        }

        return discoverySession;
      }),
    ).pipe(
      map((discoverySession) => {
        if (null === discoverySession) {
          return;
        }
        store.dispatch(new DiscoveryRequestStore.RequestDiscoverStop({ discoverySession }));
      }),
    ).subscribe();
  }

  init(discoverRequestInfo: DiscoverRequestInfo) {
    this._discoverRequestInfo = discoverRequestInfo;
    this.discoverySubject = new Subject();
    this.probeService.subscribeNotification(this);
  }

  destroy(requireDispatch: boolean = true) {
    this.probeService.unsubscribeNotification(this);

    if (requireDispatch) {
      this.store.dispatch(new DiscoveryRequestStore.DiscoveryComplete());
    }

    this._requestID = null;
    this.startDate = null;

    this._discoverRequestInfo = null;
    this.discoverySubject = null;
  }

  public get requestID(): string {
    return this._requestID;
  }

  public get discoverRequestInfo(): DiscoverRequestInfo {
    return this._discoverRequestInfo;
  }

  public discoveryObservable(): Observable<{ messageType: DiscoveryMessageType, params: any[] }> {
    return this.discoverySubject.asObservable();
  }


  /**
   * DiscoveryQueueing
   */
  @RPCSubscriber({ method: 'DiscoveryService.Queueing' })
  public DiscoveryQueueing(requestID: string, queueingDate: Date) {
    this._requestID = requestID;

    this.store.dispatch(new DiscoveryRequestStore.ChangeDiscoveryStatus({ discoveryStatus: DiscoveryStatusType.Queueing }));
    this.discoverySubject.next({ messageType: DiscoveryMessageType.Queueing, params: [requestID, queueingDate] });
  }

  /**
   * DiscoveryQueueingFailed
   */
  @RPCSubscriber({ method: 'DiscoveryService.QueueingFailed' })
  public DiscoveryQueueingFailed(queueingFailedDate: Date) {
    this.store.dispatch(new DiscoveryRequestStore.ChangeDiscoveryStatus({ discoveryStatus: DiscoveryStatusType.Stopped }));
    this.discoverySubject.next({ messageType: DiscoveryMessageType.QueueingFailed, params: [queueingFailedDate] });

    this.destroy();
  }

  /**
   * DiscoveryQueueingTimeout
   */
  @RPCSubscriber({ method: 'DiscoveryService.QueueingTimeout' })
  public DiscoveryQueueingTimeout(queueingTimeoutDate: Date) {
    this._requestID = null;

    this.store.dispatch(new DiscoveryRequestStore.ChangeDiscoveryStatus({ discoveryStatus: DiscoveryStatusType.Stopped }));
    this.discoverySubject.next({ messageType: DiscoveryMessageType.QueueingTimeout, params: [queueingTimeoutDate] });

    this.destroy();
  }

  /**
   * DiscoveryStart
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveryStart' })
  public DiscoveryStart(startDate: Date) {
    this.startDate = startDate;

    this.store.dispatch(new DiscoveryRequestStore.ChangeDiscoveryStatus({ discoveryStatus: DiscoveryStatusType.Started }));
    this.discoverySubject.next({ messageType: DiscoveryMessageType.DiscoveryStart, params: [startDate] });
  }

  /**
   * DiscoveryStop
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveryStop' })
  public DiscoveryStop(stopDate: Date) {
    this.store.dispatch(new DiscoveryRequestStore.ChangeDiscoveryStatus({ discoveryStatus: DiscoveryStatusType.Stopped }));
    this.discoverySubject.next({ messageType: DiscoveryMessageType.DiscoveryStop, params: [stopDate] });

    this.destroy();
  }

  /**
 * DiscoveryMode
 */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveryMode' })
  public DiscoveryMode(discoveryMode: DiscoveryModeType) {
    this.store.dispatch(new DiscoveryRequestStore.ChangeDiscoveryMode({ discoveryMode: discoveryMode }));
    this.discoverySubject.next({ messageType: DiscoveryMessageType.DiscoveryMode, params: [discoveryMode] });
  }

  /**
   * DiscoveryError
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveryError' })
  public DiscoveryError(err: RPCError) {
    this.discoverySubject.next({ messageType: DiscoveryMessageType.DiscoveryError, params: [err] });
  }

  /**
   * DiscoveredHost
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveredHost' })
  public DiscoveredHost(host: Host) {
    this.discoverySubject.next({ messageType: DiscoveryMessageType.DiscoveredHost, params: [host] });
  }

  /**
   * DiscoveredPort
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveredPort' })
  public DiscoveredPort(port: Port) {
    this.discoverySubject.next({ messageType: DiscoveryMessageType.DiscoveredPort, params: [port] });
  }

  /**
   * DiscoveredService
   */
  @RPCSubscriber({ method: 'DiscoveryService.DiscoveredService' })
  public DiscoveredService(service: Service) {
    this.discoverySubject.next({ messageType: DiscoveryMessageType.DiscoveredService, params: [service] });
  }

}
